@@ -2,11 +2,15 @@ from datetime import datetime
 from functools import cached_property
 from json import dumps
 from json import loads
+from os import listdir
 from os import makedirs
 from os.path import abspath
 from os.path import dirname
+from os.path import isdir
+from os.path import isfile
 from os.path import join
 from re import sub
+from shutil import copy
 from sqlite3 import Connection
 from sqlite3 import Cursor
 from sqlite3 import DatabaseError
@@ -24,7 +28,6 @@ from psutil import Process
 from psutil import process_iter
 
 from .__version__ import __version__
-from .merge import merge_database
 from .tables import journals_table
 from .tables import list_columns
 from .tables import make_journals_table
@@ -61,25 +64,65 @@ def tiered_path(id_: Union[int, str], depth: int = 5, width: int = 2) -> str:
     return join(*[id_str[n:n + width] for n in range(0, depth * width, width)])
 
 
+def copy_folder(src: str, dest: str):
+    if isdir(src):
+        for item in listdir(src):
+            copy_folder(join(src, item), join(dest, item))
+    elif isfile(src) and not isfile(dest):
+        makedirs(dirname(dest), exist_ok=True)
+        copy(src, dest)
+
+
 def clean_username(username: str) -> str:
     return str(sub(r"[^a-zA-Z0-9\-.~,]", "", username.lower().strip()))
 
 
+def copy_cursors(db_a: 'FADatabase', db_b: 'FADatabase', *cursors: 'FADatabaseCursor', replace: bool = True):
+    """
+    A -> B
+    """
+    assert all(c.table.database.database_path == db_a.database_path for c in cursors), \
+        "Cursors must point to the database calling copy"
+    assert all(set(c.columns) == set(c.table.columns) for c in cursors), "Cursors must contain all table columns"
+    db_a.check_version(patch=False, version=db_b.version)
+    db_a.check_version(patch=False)
+
+    db_a_files_folder: str = join(dirname(db_a.database_path), db_a.settings["FILESFOLDER"])
+    db_b_files_folder: str = join(dirname(db_b.database_path), db_b.settings["FILESFOLDER"])
+
+    for cursor in cursors:
+        table_b: FADatabaseTable = db_b[cursor.table.table]
+        for entry in cursor:
+            if not replace and table_b.table == users_table and (entry_b := table_b[entry[table_b.column_id]]):
+                entry["FOLDERS"] = list(set(entry["FOLDERS"] + entry_b["FOLDERS"]))
+            if table_b.table == submissions_table:
+                copy_folder(join(db_a_files_folder, p := tiered_path(entry["ID"])), join(db_b_files_folder, p))
+            table_b.insert(cursor.table.format_dict(entry), replace=True)
+
+
+def format_list(obj: list[Value]) -> str:
+    return "".join(f"|{e}|" for e in sorted(set(map(str, obj)), key=str.lower))
+
+
+def unpack_list(obj: str) -> list[str]:
+    return [e for e in obj.split("|") if e]
+
+
 class FADatabaseTable:
     def __init__(self, database: 'FADatabase', table: str):
         self.database: 'FADatabase' = database
         self.table: str = table.upper()
-        self.list_columns: list[str] = list_columns.get(self.table, [])
+        self.list_columns: list[str] = list(map(str.upper, list_columns.get(self.table, [])))
 
     def __len__(self) -> int:
         return self.database.connection.execute(f"SELECT COUNT(*) FROM {self.table}").fetchone()[0]
 
     def __getitem__(self, key: Union[Key, Entry]) -> Optional[Entry]:
         key = key if isinstance(key, dict) else {self.column_id: key}
-        return entry[0] if (entry := list(self.cursor_to_dict(self.select(key)))) else None
+        return entry[0] if (entry := list(self.select(key))) else None
 
     def __setitem__(self, key: Key, values: Entry):
-        values = {k.upper(): self.format_list(v) if isinstance(v, list) else v for k, v in values.items()}
+        values = self.format_dict(values)
         values[self.column_id] = key
         self.insert(values)
 
@@ -90,12 +133,12 @@ class FADatabaseTable:
         return self[key] is not None
 
     def __iter__(self) -> Generator[Entry, None, None]:
-        return self.cursor_to_dict(self.select())
+        return self.select().__iter__()
 
     @cached_property
-    def columns_info(self) -> list[tuple[str, str]]:
+    def columns_info(self) -> list[tuple[str, ...]]:
         return [
-            info[1:]
+            (info[1].upper(), *info[2:])
             for info in self.database.connection.execute(f"pragma table_info({self.table})")
         ]
 
@@ -117,7 +160,7 @@ class FADatabaseTable:
             return False
         item = {k: item[k] for k in values.keys()}
         item_new = {k: sorted(filter(bool, set(item[k] + v)), key=str.lower) for k, v in values.items()}
-        self.update({k: self.format_list(v) for k, v in item_new.items()}, key) if item_new != item else None
+        self.update({k: format_list(v) for k, v in item_new.items()}, key) if item_new != item else None
         return item_new != item
 
     def remove_from_list(self, key: Key, values: dict[str, list[Value]]) -> bool:
@@ -127,7 +170,7 @@ class FADatabaseTable:
             return False
         item = {k: item[k] for k in values.keys()}
         item_new = {k: sorted(filter(bool, set(item[k]) - set(v)), key=str.lower) for k, v in values.items()}
-        self.update({k: self.format_list(v) for k, v in item_new.items()}, key) if item_new != item else None
+        self.update({k: format_list(v) for k, v in item_new.items()}, key) if item_new != item else None
         return item_new != item
 
     def reload(self):
@@ -135,33 +178,20 @@ class FADatabaseTable:
         for attr in filter(self.__dict__.__contains__, ("columns", "columns_info", "column_id")):
             del self.__dict__[attr]
 
-    def cursor_to_dict(self, cursor: Cursor, columns: list[str] = None) -> Generator[Entry, None, None]:
-        columns = list(map(str.upper, self.columns if columns is None else columns))
-        return ({k: self.unpack_list(v) if k in self.list_columns else v for k, v in zip(columns, entry)}
-                for entry in cursor)
-
-    @staticmethod
-    def format_list(obj: list[Value]) -> str:
-        return "".join(f"|{e}|" for e in sorted(set(map(str, obj)), key=str.lower))
-
-    @staticmethod
-    def unpack_list(obj: str) -> list[str]:
-        return [e for e in obj.split("|") if e]
-
     def format_dict(self, obj: Entry) -> dict[str, Value]:
         obj = {k.upper().replace("_", ""): v for k, v in obj.items()}
-        obj = {k: self.format_list(v) if isinstance(v := obj.get(k, ""), list) else v for k in self.columns}
+        obj = {k: format_list(v) if isinstance(v := obj.get(k, ""), list) else v for k in self.columns}
         return obj
 
     def select(self, query: dict[str, Union[list[Value], Value]] = None, columns: list[str] = None,
                query_and: bool = True, query_and_values: bool = False, like: bool = False,
                order: list[str] = None, limit: int = 0, offset: int = 0
-               ) -> Cursor:
-        query = {} if query is None else query
+               ) -> 'FADatabaseCursor':
+        query = query or {}
         query = {k: [v] if not isinstance(v, list) else v for k, v in query.items()}
         query = {k: vs for k, vs in query.items() if vs}
-        order = [] if order is None else order
-        columns = self.columns if columns is None else columns
+        order = order or []
+        columns = columns or self.columns
         op: str = "like" if like else "="
         logic: str = "AND" if query_and else "OR"
         logic_values: str = "AND" if query_and_values else "OR"
@@ -170,7 +200,7 @@ class FADatabaseTable:
             for k, vs in query.items()
         ]))
         order_str = ",".join(order)
-        return self.database.connection.execute(
+        cursor: Cursor = self.database.connection.execute(
             f"""SELECT {','.join(columns)} FROM {self.table}
             {f' WHERE {where_str} ' if where_str else ''}
             {f' ORDER BY {order_str} ' if order_str else ''}
@@ -178,12 +208,13 @@ class FADatabaseTable:
             {f' OFFSET {offset} ' if limit > 0 and offset > 0 else ''}""",
             [v for values in query.values() for v in values]
         )
+        return FADatabaseCursor(cursor, columns, self)
 
     def select_sql(self, where: str = "", values: list[Value] = None, columns: list[str] = None,
-                   order: list[str] = None, limit: int = 0, offset: int = 0) -> Cursor:
-        columns = self.columns if columns is None else columns
-        order = [] if order is None else order
-        return self.database.connection.execute(
+                   order: list[str] = None, limit: int = 0, offset: int = 0) -> 'FADatabaseCursor':
+        columns = columns or self.columns
+        order = order or []
+        cursor: Cursor = self.database.connection.execute(
             f"""SELECT {','.join(columns)} FROM {self.table}
             {f' WHERE {where} ' if where else ''}
             {f' ORDER BY {",".join(order)}' if order else ''}
@@ -191,6 +222,7 @@ class FADatabaseTable:
             {f' OFFSET {offset}' if limit > 0 and offset > 0 else ''}""",
             [] if values is None else values
         )
+        return FADatabaseCursor(cursor, columns, self)
 
     def insert(self, values: dict[str, Value], replace: bool = True):
         self.database.connection.execute(
@@ -229,10 +261,10 @@ class FADatabaseJournals(FADatabaseTable):
 
 class FADatabaseSettings(FADatabaseTable):
     def __getitem__(self, setting: str) -> Optional[str]:
-        return entry["SVALUE"] if (entry := super().__getitem__(setting)) is not None else None
+        return (super().__getitem__(setting) or {}).get("SVALUE", None)
 
     def __setitem__(self, setting: str, value: str):
-        self.insert({"SETTING": setting, "SVALUE": value})
+        super().__setitem__(setting, {"SVALUE": value})
 
     def read_history(self) -> list[tuple[float, str]]:
         return list(map(tuple, loads(h) if (h := self["HISTORY"]) else []))
@@ -293,20 +325,19 @@ class FADatabaseSubmissions(FADatabaseTable):
 
 class FADatabaseUsers(FADatabaseTable):
     def new_user(self, user: str) -> str:
-        user = clean_username(user)
-        if user not in self:
-            self[user] = {f: "" for f in self.columns}
+        if (user := clean_username(user)) not in self:
+            self[user] = {}
         return user
 
     def activate_user(self, user: str):
         if (user_entry := self[(user := clean_username(user))]) is None:
             return
-        self.update({"FOLDERS": self.format_list([f.strip("!") for f in user_entry["FOLDERS"]])}, user)
+        self.update({"FOLDERS": format_list([f.strip("!") for f in user_entry["FOLDERS"]])}, user)
 
     def deactivate_user(self, user: str):
         if (user_entry := self[(user := clean_username(user))]) is None:
             return
-        self.update({"FOLDERS": self.format_list([f"!{f.strip('!')}" for f in user_entry["FOLDERS"]])}, user)
+        self.update({"FOLDERS": format_list([f"!{f.strip('!')}" for f in user_entry["FOLDERS"]])}, user)
 
     def add_user_folder(self, user: str, folder: str):
         if not (user_entry := self[(user := clean_username(user))]):
@@ -321,6 +352,18 @@ class FADatabaseUsers(FADatabaseTable):
         self.remove_from_list(clean_username(user), {"FOLDERS": [folder.lower()]})
 
 
+class FADatabaseCursor:
+    def __init__(self, cursor: Cursor, columns: list[str], table: FADatabaseTable):
+        self.table: FADatabaseTable = table
+        self.cursor: Cursor = cursor
+        self.columns: list[str] = columns
+
+    def __iter__(self) -> Generator[Entry, None, None]:
+        columns = list(map(str.upper, self.columns))
+        return ({k: unpack_list(v) if k in self.table.list_columns else v for k, v in zip(columns, entry)}
+                for entry in self.cursor)
+
+
 class FADatabase:
     def __init__(self, database_path: str):
         self.database_path: str = abspath(database_path)
@@ -389,14 +432,15 @@ class FADatabase:
         self.users.reload()
         self.committed_changes: int = self.total_changes
 
-    def check_version(self, raise_for_error: bool = True, *, major: bool = True, minor: bool = True, patch: bool = True
+    def check_version(self, raise_for_error: bool = True, *, major: bool = True, minor: bool = True, patch: bool = True,
+                      version: str = __version__
                       ) -> Optional[DatabaseError]:
         err: Optional[DatabaseError] = None
-        if (v := self.version) == __version__:
+        if (v := self.version) == version:
             return None
         if not v:
             err = DatabaseError(f"version error: {v}")
-        if (v_db := v.split("."))[0] != (v_md := __version__.split("."))[0]:
+        if (v_db := v.split("."))[0] != (v_md := version.split("."))[0]:
             err = DatabaseError(f"major version is not latest: {v_db[0]} != {v_md[0]}") if major else None
         elif v_db[1] != v_md[1]:
             err = DatabaseError(f"minor version is not latest: {v_db[1]} != {v_md[1]}") if minor else None
@@ -410,12 +454,12 @@ class FADatabase:
         ps: list[Process] = []
         for process in process_iter():
             try:
-                if process.is_running() and any(self.database_path == path for path, _fd in process.open_files()):
+                if process.is_running() and any(self.database_path == f.path for f in process.open_files()):
                     ps.append(process)
+                if len(ps) > 1 and raise_for_error:
+                    raise Error(f"Multiple connections to database: {ps}")
             except (NoSuchProcess, AccessDenied):
                 pass
-        if len(ps) > 1 and raise_for_error:
-            raise Error(f"Multiple connections to database: {ps}")
         return ps
 
     def commit(self):
@@ -432,8 +476,12 @@ class FADatabase:
     def close(self):
         self.connection.close()
 
-    def merge(self, db_b: 'FADatabase'):
-        merge_database(self.connection, dirname(self.database_path), db_b.connection, dirname(db_b.database_path))
+    def merge(self, db_b: 'FADatabase', *cursors: FADatabaseCursor):
+        copy_cursors(db_b, self, *cursors or (db_b.users.select(), db_b.submissions.select(), db_b.journals.select()),
+                     replace=False)
+
+    def copy(self, db_b: 'FADatabase', *cursors: FADatabaseCursor):
+        copy_cursors(self, db_b, *cursors or (self.users.select(), self.submissions.select(), self.journals.select()))
 
     def vacuum(self):
         self.connection.execute("VACUUM")
