@@ -1,7 +1,5 @@
 from json import dumps
 from sqlite3 import Connection
-from typing import Dict
-from typing import List
 
 from .__version__ import __version__
 
@@ -48,7 +46,7 @@ settings_table: str = "SETTINGS"
 submissions_table: str = "SUBMISSIONS"
 users_table: str = "USERS"
 
-list_columns: Dict[str, List[str]] = {
+list_columns: dict[str, list[str]] = {
     journals_table: ["MENTIONS"],
     settings_table: [],
     submissions_table: ["TAGS", "FAVORITE", "MENTIONS"],
@@ -104,7 +102,7 @@ def make_submissions_table(db: Connection):
         TYPE TEXT NOT NULL CHECK (TYPE IN ('image', 'music', 'text', 'flash')),
         FILEURL TEXT NOT NULL,
         FILEEXT TEXT NOT NULL,
-        FILESAVED INT NOT NULL CHECK (FILESAVED in (0, 1)),
+        FILESAVED INT NOT NULL CHECK (FILESAVED in (0, 1, 10, 11)),
         FAVORITE TEXT NOT NULL,
         MENTIONS TEXT NOT NULL,
         FOLDER TEXT NOT NULL CHECK (FOLDER IN ('gallery', 'scraps')),
